#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <atari.h>
#include "support.h"

/* Bitmap lookup tables
 *                                                                           /--GTIA--\
 *               ANTIC mode:   2,  3,  4,  5,  6,  7,  8,  9,  A,  B,  C,  D,  E,  F  10  11  12  MCS(13)
 */
const unsigned char _ppb[] = { 1,  1,  1,  1,  1,  1,  4,  8,  4,  8,  8,  4,  4,  8,  2,  2,  2,  4};
const unsigned char _bak[] = { 0,  0,  0,  0,  0,  0,  4,  4,  4,  4,  4,  4,  4,  4,  0,  0,  0,  4};

#define BYTES_PER_LINE(bmp) (bmp->w / PIXELS_PER_BYTE(bmp->mode))
#define PIXELS_PER_BYTE(x)  _ppb[x-2]
#define BAKCOL(x)           _bak[x-2]

const unsigned char bmpclear_4c_mask[3] = {0x55, 0xAA, 0xFF};

//*** Antic mode instruction bits ***//
#define ANTIC_DLI   128
#define ANTIC_LMS   64
#define ANTIC_VSCRL 32
#define ANTIC_HSCRL 16

#define MAX_DL_LEN      1024

#define SDMCTL  *(unsigned char *)  0x022F          // dma control shadow
#define SDLSTL  *(unsigned int *)   0x0230          // DL address shadow

const unsigned char plot4cmask[4]  = {0x3F,0xCF,0xF3,0xFC};
const unsigned char plot4cshift[4] = {6,4,2,0};

/* atascii_to_internal
 *
 * Converts an atascii character set value to an internal character set
 * value.
 */
char __fastcall__ atascii_to_internal(char c)
{
  if (c < 32)
    return(c + 64);
  else if (c < 96)
    return(c - 32);
  else
    return(c);
}


/* aligned_malloc
 *
 * Allocates an aligned block of memory of size, aligned to boundary
 * that you pass in.  Returns pointer to the memory block or NULL if
 * error allocating memory.
 */
void* __fastcall__ aligned_malloc(size_t size, unsigned int bound)
{
  char *d, *b;                          //           |
  unsigned int lo_free;                 //           |
  unsigned int hi_free;                 //           |
  unsigned int bnd;


  bnd = bound - 1;


  //*** Allocate an aligned block of memory ***//
  d = (char *) malloc(size + bnd);                          // allocate mem
  if (!d) return(NULL);                                     // mem not avail

  b = (char *) ((((unsigned int) d) + bnd) & ~bnd);         // aligned pointer

  lo_free = b - d;                                          // memory not used
  hi_free = bnd - lo_free;                                  //       |
  _heapadd(d, lo_free);                                         // free low
  _heapadd(b + bound, hi_free);                                 // free high

  return(b);
}


/* plot_4colmode_nc
 *
 * Plots a point on the bitmap using the passed in color register.  This
 * function does no clipping, so you have to watch the parameters you pass
 * in.
 */
void __fastcall__ plot_4colmode_nc(BITMAP *bmp, unsigned int x, unsigned int y, unsigned char creg)
{
  unsigned char bit;
  unsigned char data;
  unsigned int index;


  bit  = (x % 4);
  index = (bmp->bpl * y) + (x / 4);

  data = bmp->dat[index];
  data &= plot4cmask[bit];                      // clear the pixel first

  if (creg != BAKCOL(bmp->mode)) {              // not background color?
    data |= ((creg+1) << plot4cshift[bit]);     // creg 0->01, 1->10, 2->11
  }

  bmp->dat[index] = data;
  return;
}


/* create_bitmap
 *
 * Creates a bitmap for drawing or printing text to.  Pass in the antic
 * mode of this bitmap and the height and width.  Returns a pointer to
 * a BITMAP structure, or NULL if some error.  Free the memory with the
 * destroy_bitmap function.
 */
BITMAP* __fastcall__ create_bitmap(unsigned char mode, unsigned int w, unsigned int h)
{
  char *b, *d;
  unsigned int lo_free;
  unsigned int hi_free;
  unsigned int bytes;
  unsigned int extra;
  BITMAP *bmp;


  //*** Allocate memory for bitmap structure ***//
  bmp = (BITMAP *) malloc(sizeof(BITMAP));
  if (!bmp) return(NULL);

  bmp->mode = mode;
  bmp->w    = w;
  bmp->h    = h;
  bmp->bpl  = BYTES_PER_LINE(bmp);
  //bmp->clip = 1;

  bytes = (bmp->bpl * h);                           // total bytes for buffer
  if (bytes > 4096) {
    extra = (unsigned char) (4096 / bmp->bpl);      // # of screen lines fit
    extra = (unsigned char) (4096 - (bmp->bpl * extra));
  }
  else
    extra = 0;

  //*** Align on a 4k boundary if bitmap less than 4k ***//
  d = (char *) malloc((bytes) + 0xFFF + extra);          // allocate mem
  b = (char *) ((((unsigned int) d) + 0xFFF) & ~0xFFF);  // aligned pointer
  b = b + extra;                                         // add extra

  lo_free = b - d;                                       // memory not used
  hi_free = 0xFFF - lo_free;                             //       |
  _heapadd(d, lo_free);                                      // free low chunk
  _heapadd(b + (bytes + extra), hi_free);                    // free high chunk

  bmp->dat = b;
  if (!bmp->dat) {
    free(bmp);
    return(NULL);
  }

  return(bmp);
}



/* clear_bitmap_to
 *
 * Clear a bitmap to a specific color (color register 0-4)
 */
void __fastcall__ clear_bitmap_to(BITMAP *bmp, unsigned char creg)
{
  unsigned int bytes;
  unsigned char c;



  bytes = (bmp->bpl * bmp->h);

  switch(PIXELS_PER_BYTE(bmp->mode)) {
    case 1:
      c = creg;
      break;

    case 2:
      c = (creg | (creg << 4));
      break;

    case 4:
      if (creg == 4)
	c = 0;
      else
	c = bmpclear_4c_mask[creg];
      break;

    case 8:
      if (creg == 4)
	c = 0;
      else
	c = 0xFF;
      break;
  }

  memset(bmp->dat, c, bytes);
  return;
}


/* clear_bitmap
 *
 * Clears the bitmap to the background color.
 */
void __fastcall__ clear_bitmap(BITMAP *bmp)
{
  clear_bitmap_to(bmp, BAKCOL(bmp->mode));
  return;
}

/* destroy_bitmap
 *
 * Frees all the memory being used by the bitmap.
 */
BITMAP* __fastcall__ destroy_bitmap(BITMAP *bmp)
{
  unsigned int bytes;


  bytes = (bmp->bpl * bmp->h);

  _heapadd(bmp->dat, bytes);
  free(bmp);

  return(NULL);
}




/* create_display_list
 *
 * Creates the display list from the display list structure, allocates
 * memory for it (on a 1k boundary) and returns the address to the user.
 * The display list is not active until the user activates it.
 */
char* __fastcall__ create_display_list(DL_SETUP_T *list)
{
  char *dl;                             // temporary DL storage
  unsigned char *t;                      // aligned memory pointers
  char *b;                              //           |
//  unsigned int lo_free;                 //           |
//  unsigned int hi_free;                 //           |
  int m;                                // dl mode instruction
  int i, n, l;                          // counters
  unsigned int bytes;                   // bytes to line


  //*** Initialize ***//
  i = 0;                                // location in temp array
  n = 0;                                // display list setup entry
  dl = (char *) malloc(MAX_DL_LEN);     // allocate temp storage for dl
  if (!dl) return(NULL);

  //*** Build display list ***//
  do {
    if (list[n].mode == 0xFF) break;                      // last setup entry

    if ((list[n].mode < 0x10) && (list[n].mode > 1)) {    // antic mode entry
      //*** Deal with first line ***//
      m = list[n].mode;

      if (list[n].hscroll)  m += ANTIC_HSCRL;             // horiz fine scroll
      if (list[n].vscroll)  m += ANTIC_VSCRL;             // vert fine scroll

      if ((list[n].dli_flag) && (list[n].lines == 1))
	 m += ANTIC_DLI;                                  // DLI flag

      if (list[n].bmp) {                                  // Load memory scan
	m += ANTIC_LMS;
	dl[i] = m;
	i++;

	t = (unsigned char *) &dl[i];                      // LMS address
	*t++ = (unsigned char) ((unsigned int)list[n].bmp->dat & 0xFF);             //     |
	*t = (unsigned char) ((unsigned int)list[n].bmp->dat >> 8);             //     |
	i += 2;
      }

      //*** Deal with rest of the lines ***//
      for(l=1; l < list[n].lines; l++) {
	m = list[n].mode;                                 // mode

	if ((list[n].dli_flag) && (l == list[n].lines-1)) // dli on last line
	  m += ANTIC_DLI;

	if (list[n].vscroll)  m += ANTIC_VSCRL;           // vert fine scroll

	bytes = (unsigned int) (l * (list[n].bmp->bpl));

	if (list[n].hscroll) {                            // horiz fine scroll
	  m += ANTIC_HSCRL;
	  m += ANTIC_LMS;
	  dl[i] = m;
	  i++;

	  t = (unsigned char *) &dl[i];                      // LMS address
	*t++ = (unsigned char) (((unsigned int)list[n].bmp->dat+bytes) & 0xFF);             //     |
	*t = (unsigned char) (((unsigned int)list[n].bmp->dat+bytes) >> 8);             //     |
	  i += 2;

	  continue;
	}

	//*** 4k boundary LMS instruction? ***//
	//lo_free = (unsigned int) (4096 / (list[n].bmp->bpl);

	if ((((unsigned int)list[n].bmp->dat + bytes) % 4096) == 0) {
	  m += ANTIC_LMS;
	  dl[i] = m;
	  i++;

	  t = (unsigned char *) &dl[i];                      // LMS address
	*t++ = (unsigned char) (((unsigned int)list[n].bmp->dat+bytes) & 0xFF);             //     |
	*t = (unsigned char) (((unsigned int)list[n].bmp->dat+bytes) >> 8);             //     |
	  i += 2;

	  continue;
	}

	dl[i] = m;
	i++;
      }
    }
    else {
      dl[i] = list[n].mode;             // blank line statements
      i++;
    }

    n++;                                // next line in dl setup list
  } while (1);

  //*** Allocate 1k aligned block for display list ***//
  b = aligned_malloc(i+4, 1024);            // 1k aligned block

  //*** Jump and wait for vblank instruction ***//
  dl[i] = 65;
  i++;
	  t = (unsigned char *) &dl[i];                 // JVB address (start of list)
	*t++ = (unsigned char) ((unsigned int)b & 0xFF);             //     |
	*t = (unsigned char) ((unsigned int)b >> 8);             //     |
  i += 2;                                   // dl array index

  memcpy(b, dl, i);                         // copy the dl into allocd mem
  free(dl);                                 // free our temporary array
  return(b);
}

/* destroy_display_list
 *
 * Frees all memory associated with the display list.
 */
void __fastcall__ destroy_display_list(char *b)
{
  int i;


  //*** Count bytes in display list ***//
  for(i=0; b[i] != 65; i++);
  i += 3;                           // account for jvb address + 1

  _heapadd(b, i);                       // free memory
  return;
}


/* activate_display_list
 *
 * Activates the display list pointed to by dl.  Saves the current state
 * of 559, sets SDLSTL (560, 561) and turns Antic back on.
 */
void __fastcall__ activate_display_list(char *dl)
{
  SDMCTL -= 32;                     // turn off antic dma
  SDLSTL = (unsigned int) dl;       // assign new display list address
  SDMCTL += 32;                     // turn on antic dma

  return;

}



/* textout_4colmode
 *
 * Writes text to the bitmap without clipping.  Use this function with
 * the four color bitmap modes (Antic 8, A, D), and make sure you are not
 * clobbering memory beyond your bitmap.  Color is the color register to
 * use for plotting the text.
 *
 * Font is a pointer to a font in memory, and can be the system font.
 */
void __fastcall__ textout_4colmode_nc(BITMAP *bmp, FONT *font, const char *s, unsigned int x, unsigned int y, char creg)
{
  unsigned char i;                      // string counter
  unsigned char n, m;                   // byte and bit counters
  unsigned int tx, ty, sx;              // screen position
  char c;                               // character
  unsigned char *line;                  // character line pointer in font


  //*** Initialize ***//
  tx   = x;
  sx   = tx;

  for(i=0; i<strlen(s); i++) {
    c = atascii_to_internal(s[i]);                // convert to internal
    line = font + (c * 8);                        // offset to font data
    ty = y;

    for (n=0; n<8; n++) {                         // each line in character
      for (m=128; m > 1; m = m / 2) {             // each bit in line
	if (*line & m)
	  plot_4colmode_nc(bmp, tx, ty, creg);
	tx++;
      }

      tx = sx;                                  // start of char on screen
      ++line;                                   // next line of character
      ++ty;                                     // next screen line down
    }

    sx += 8;                                    // next char pos on screen
    tx = sx;
  }

  return;
}



/*
   Assembler routine move shapes in screen
   n1 -> Memory address destination in screen
   n2 -> Memory address shape source
   n3 -> #bytes per row
   n4 -> #rows shape
   n5 -> #bytes/row in screen
*/
void mover(int n1, unsigned char *n2, int n3, int n4, int n5)
{

   pokew(0xCB,n1);
   pokew(0xCD,n2);
   poke(0xCF,n3);   // #rows shape
   poke(0xD2,n4);   // pass to registro X   (nrows)
   poke(0xD0,n5);   // #bytes/row in screen

   asm("  LDX $D2");       // Load register X
   asm("_label4:");
   asm("  LDY $CF");
   asm("  DEY");
   asm("_label1:");
   asm("  LDA ($CD),Y");
   asm("  STA ($CB),Y");
   asm("  DEY");
   asm("  CPY #$FF");
   asm("  BNE _label1");
   asm("  CLC");
   asm("  LDA $CB");
   asm("  ADC $D0");
   asm("  STA $CB");
   asm("  BCC _label2");
   asm("  INC $CC");
   asm("_label2:");
   asm("  CLC");
   asm("  LDA $CD");
   asm("  ADC $CF");
   asm("  STA $CD");
   asm("  BCC _label3");
   asm("  INC $CE");
   asm("_label3:");
   asm("  DEX");
   asm("  BNE _label4");

}

/*
   Assembler routine copy the same char in a block memory
   n1 -> Start address
   n2 -> #bytes
   b3 -> char
*/
void copia_char(unsigned char *n1, int n2, unsigned char n3)
{

   pokew(0xCB,n1);  // Start address
   pokew(0xCD,n2);  // #bytes
   poke(0xCF,n3);   // Char

   // Works with Hi-byte first
   asm("   LDX $CE");
   asm("   CPX #$00"); // leave if hi-byte is zero
   asm("   BEQ _label7");
   asm("   LDY #$00");
   asm("_label5:");
   asm("   DEY");
   asm("   CLC");
   asm("_label6:");
   asm("   LDA $CF");
   asm("   STA ($CB),Y");
   asm("   DEY");
   asm("   CPY #$FF");
   asm("   BNE _label6");
   asm("   CLC");
   asm("   INC $CC");
   asm("   CLC");
   asm("   LDY #$00");
   asm("   DEX");
   asm("   BNE _label5");

   // now works with low-byte
   asm("_label7:");
   asm("   LDY $CD");
   asm("   DEY");
   asm("_label8:");
   asm("   LDA $CF");
   asm("   STA ($CB),Y");
   asm("   DEY");
   asm("   CPY #$FF");
   asm("   BNE _label8");
   asm("   CLC");

}

/*
   Assembler routine copy the same char in a shape block memory in screen
   n1 -> Start address shape screen
   n2 -> Char
   n3 -> #bytes/row shape
   n4 -> #rows
   n5 -> #bytes/row screen
*/
void copia_char_bloque(int n1, int n2, int n3, int n4, int n5)
{

   pokew(0xCB,n1);  // Start address shape screen
   pokew(0xCD,n2);  // Char
   poke(0xCF,n3);   // #bytes/row shape
   poke(0xD2,n4);   // #rows
   poke(0xD0,n5);   // #bytes/row screen

   asm("  LDX $D2");
   asm("_label11:");
   asm("  LDY $CF");
   asm("  DEY");
   asm("_label9:");
   asm("  LDA $CD");
   asm("  STA ($CB),Y");
   asm("  DEY");
   asm("  CPY #$FF");
   asm("  BNE _label9");
   asm("  CLC");
   asm("  LDA $CB");
   asm("  ADC $D0");
   asm("  STA $CB");
   asm("  BCC _label10");
   asm("  INC $CC");
   asm("_label10:");
   asm("  CLC");
   asm("  DEX");
   asm("  BNE _label11");

}
